"""
Plot module associated to the Biomechzoo toolbox

Example:
    from biomechzoo.visualization.ensembler import Ensembler
    ens = Ensembler(fld=bmech.in_folder, ch=["gy_shank], conditions=["pre, post], subj_pattern = [r"\b\d{3}[A-Z]{2}\b"])
    ens.combine()
"""

import numpy as np
import os
import re
import pandas as pd
import json

import plotly.graph_objects as go
from plotly.subplots import make_subplots
import plotly.colors as pc
from dash import Dash, dcc, html, Input, Output, State, no_update

from biomechzoo.utils.engine import engine
from biomechzoo.utils.zload import zload
from biomechzoo.utils.findfield import findfield

class Ensembler:
    def __init__(self, fld, ch, conditions, out_folder=None, name_contains=None, show_legend=True, match_all=True, subj_pattern=None):
        if isinstance(subj_pattern,str):
            subj_pattern = [subj_pattern]

        self.fld = fld
        self.conditions = conditions
        self.channels = ch
        self.out_folder = out_folder
        self.show_legend = show_legend
        self.subj_pattern = subj_pattern
        self.zoo_files = engine(fld, extension=".zoo", subfolders=conditions, name_contains=name_contains, match_all=match_all)
        self.fig = self._create_subplots()
        self.subject_colors = self._assign_subject_colors()

    def _assign_subject_colors(self):
        """Creates subject specific colors"""
        unique_subjects = self._get_unique_subjects()
        subject_colors = {}
        for idx, subj in enumerate(unique_subjects):
            line_color, shade_color, marker_color = self._assign_colors(idx)
            subject_colors[subj] = {
                "line": line_color,
                "shade": shade_color,
                "event": marker_color
            }
        return subject_colors

    def _get_unique_subjects(self):
        """Extract unique subject names from subject pattern initialized in __init__()"""
        # TODO: get an option when subj_pattern is None.
        #  Get from biomechzoo zoosystem?

        subjects = set()
        for fl in self.zoo_files:
            match = re.search(self.subj_pattern[0], fl)
            if match:
                subjects.add(match.group(0))
            elif match is None:
                match = re.search(self.subj_pattern[1], fl)
                if match:
                    subjects.add(match.group(0))
                else:
                    subjects.add("unknown")
        return sorted(subjects)

    @staticmethod
    def  _assign_colors(i, color_library=None):
        """
        Assign colors to each subject automatically.

        Parameters
        ----------
            i: integer
                The index associated with the subject pattern

        Returns
        --------
            hex_code: string
                The ith hex-code from pc.qualitative.D3 library.
            shade_color: string
                The associated shade color

            marker_color: string
                The complementary marker color
        """
        if color_library is None:
            color_library = pc.qualitative.D3

        hex_code = color_library[i % len(color_library)]
        h = hex_code.lstrip('#')
        rgb =tuple(int(h[i:i + 2], 16) for i in (0, 2, 4))

        #shade color with opacity
        opacity = 0.3
        shade_color = f"rgba({rgb[0]}, {rgb[1]}, {rgb[2]}, {opacity})"

        #Get complementary color for marker
        comp = ['%02X' % (255 - a) for a in rgb]
        marker_color =  '#' + ''.join(comp)

        return hex_code, shade_color, marker_color


    def _create_subplots(self):
        """Create subplots for each channel and each condition"""
        self.rows = len(self.channels)
        self.cols = len(self.conditions)
        titles = [f"{ch} - {cond}" for ch in self.channels for cond in self.conditions]
        fig = make_subplots(rows=self.rows, cols=self.cols, shared_xaxes=True, shared_yaxes=False,
                             subplot_titles=titles)
        return fig

    def _create_subplots_combine(self):
        """Create subplots for each channel"""
        self.rows = len(self.channels)
        self.cols = 1
        titles = [f"{ch}" for ch in self.channels ]
        fig = make_subplots(rows=self.rows, cols=self.cols, shared_xaxes=True, shared_yaxes=True,
                            subplot_titles=titles)
        return fig

    def _get_condition_from_path(self, path):
        for cond in self.conditions:
            if cond in path:
                return cond
        return "Unknown"

    def _make_point_customdata(self, subj, channel, condition, fname, row, col, x, y):
        """Curate data for the hover functionality in plotly figure"""
        # Ensure x is an array of indices when None
        if x is None:
            x = list(range(len(y)))

        if isinstance(y, float):
            return [
                {
                    "subject": subj,
                    "channel": channel,
                    "condition": condition,
                    "source_file": fname,
                    "row": row,
                    "col": col,
                    "index": int(x) if isinstance(x, (int, np.integer)) else x,
                    "value": float(y) if isinstance(y, (float, np.floating)) else y
                }
            ]

        return [
            {
                "subject": subj,
                "channel": channel,
                "condition": condition,
                "source_file": fname,
                "row": row,
                "col": col,
                "index": int(xi) if isinstance(xi, (int, np.integer)) else xi,
                "value": float(yi) if isinstance(yi, (float, np.floating)) else yi
            } for xi, yi in zip(x, y)
        ]

    def _default_hovertemplate(self):
        """Curate default hover template"""
        # Compact, informative hover
        return (
            "Subject: %{customdata.subject}<br>"
            "Channel: %{customdata.channel}<br>"
            "Condition: %{customdata.condition}<br>"
            "File: %{customdata.source_file}<br>"
            "x: %{x}<br>y: %{y}"
            "<extra></extra>"
        )

    def cycles(self, event_name=None):
        """
        Plot cycle data. Assumes data is normalized to 100% of the movement.

        Parameters:
        -----------
            event_name: str or list

        Returns:
        --------

        """

        # check if fig is populated
        if self.fig.data:
            self.fig.data = []

        # loop thought the zoofiles and plot the traces
        for fl in self.zoo_files:
            data = zload(fl)
            fname = os.path.basename(fl)
            condition = self._get_condition_from_path(fl)

            subj_not_found = True
            while subj_not_found:
                for key in self.subject_colors:
                    if key in fl:
                        subj = key
                        subj_not_found = False

            if subj_not_found:
                subj = "Unknown"

            line_color = self.subject_colors[subj]["line"]
            marker_color = self.subject_colors[subj]["event"]

            if not any(t.legendgroup == subj for t in self.fig.data):
                self.add_line(y=[None],name=f"Subject - {subj}", color=line_color,legendgroup=subj, showlegend=True )

            for i, channel in enumerate(self.channels):
                ch_data_line = data[channel]["line"]
                row = i + 1
                col = self.conditions.index(condition) + 1

                # Built metadata for click/hover
                x_line = list(range(len(ch_data_line)))
                cdata = self._make_point_customdata(subj, channel, condition, fname, row, col, x_line, ch_data_line)

                self.add_line(y=ch_data_line, x=x_line, row=row, col=col,
                              name=f"{fname} - {channel}", color=line_color,
                              legendgroup=subj, showlegend=False,
                              customdata=cdata, hovertemplate=self._default_hovertemplate())

                if event_name:
                    exd, eyd, evt_ch = self._get_events_data(data, event_name, fname=fname)
                    if evt_ch == channel:
                        # Create dummy marker trace for the legend
                        self.add_marker(y=[None], x=[None], name=f"{event_name} - {subj}", color=marker_color,
                                        legendgroup=subj, showlegend=True)

                        # plot the markers on the line
                        exd = np.array(exd)  # prep for plotting
                        eyd = np.array(eyd)  # prep for plotting
                        cdata_m = self._make_point_customdata(subj, channel, condition, fname, row, col, exd.tolist(),
                                                              eyd.tolist())
                        self.add_marker(y=eyd, x=exd, row=row, col=col,
                                        name=event_name, color=marker_color,
                                        legendgroup=subj, showlegend=False,
                                        customdata=cdata_m, hovertemplate=self._default_hovertemplate())

        self.show(title="Cycles per Subject")


    def quality_check_cycles(self, event_name=None):
        # check if fig is populated
        if self.fig.data:
            self.fig.data = []

        # loop thought the zoofiles and plot the traces
        for fl in self.zoo_files:
            data = zload(fl)
            fname = os.path.basename(fl)
            condition = self._get_condition_from_path(fl)

            subj_not_found = True
            while subj_not_found:
                for key in self.subject_colors:
                    if key in fl:
                        subj = key
                        subj_not_found = False

            if subj_not_found:
                subj = "Unknown"

            line_color = self.subject_colors[subj]["line"]
            marker_color = self.subject_colors[subj]["event"]

            if not any(t.legendgroup == subj for t in self.fig.data):
                self.add_line(y=[None], name=f"Subject - {subj}", color=line_color, legendgroup=subj,
                              showlegend=True)

            for i, channel in enumerate(self.channels):
                ch_data_line = data[channel]["line"]
                # A single fixed row and column number for now
                row = 1
                col = 1

                # Built metadata for click/hover
                x_line = list(range(len(ch_data_line)))
                cdata = self._make_point_customdata(subj, channel, condition, fname, row, col, x_line, ch_data_line)

                self.add_line(y=ch_data_line, x=x_line, row=row, col=col,
                              name=f"{fname} - {channel}", color=line_color,
                              legendgroup=subj, showlegend=False,
                              customdata=cdata, hovertemplate=self._default_hovertemplate())

                if event_name:
                    exd, eyd, evt_ch = self._get_events_data(data, event_name, fname)
                    if evt_ch == channel:
                        # Create dummy trace for the legend
                        self.add_marker(y=[None], x=[None], name=f"{event_name} - {subj}", color=marker_color,
                                        legendgroup=subj, showlegend=True)

                        # plot the event on the line
                        exd = np.array(exd)  # prep for plotting
                        eyd = np.array(eyd)  # prep for plotting
                        cdata_m = self._make_point_customdata(subj, channel, condition, fname, row, col, exd.tolist(),
                                                              eyd.tolist())
                        self.add_marker(y=eyd, x=exd, row=row, col=col,
                                        name=event_name, color=marker_color,
                                        legendgroup=subj, showlegend=False,
                                        customdata=cdata_m, hovertemplate=self._default_hovertemplate())


    @staticmethod
    def _get_events_data(data, target_event=None, fname=None):
        if target_event:
            evt_val, evt_ch = findfield(data,target_event)

            if not evt_ch:
                print(f" event '{target_event}' not found in any channel for file '{fname}'")
                return None, None, None

            exd = int(evt_val[0])
            eyd = float(evt_val[1])

            return exd, eyd, evt_ch

        else:
            raise ValueError(f'No target event given: ')


    def combine(self):
        # check if fig is populated
        if self.fig:
            self.fig.layout = {}
            self.fig.data = []

        # Create new figure object
        self.fig = self._create_subplots_combine()

        data = self._calculate_average()
        for c, condition in enumerate(data):
            line_color, shade_color, marker_color = self._assign_colors(c)

            if not any(t.legendgroup == condition for t in self.fig.data):
                self.add_line(y=[None],name=f"{condition}", color=line_color,legendgroup=condition, showlegend=True )

            for i, channel in enumerate(data[condition]):
                average = data[condition][channel]["average"]
                standard_dev = data[condition][channel]["standard_dev"]

                # populate the figure
                row = i + 1
                self.add_line(y=average, row=row, col=1, name=f"{condition} - {channel}", color=line_color, legendgroup=condition, showlegend=False)
                self.add_errorbar(y=average, yerr=standard_dev, row=row, col=1,
                                  color=shade_color)

        self.show()


    def combine_within(self):
        raise NotImplementedError

    def average(self):
        # check if fig is populated
        if self.fig.data:
            self.fig.data = []

        data = self._calculate_average()
        for c, condition in enumerate(data):
            line_color, shade_color, marker_color = self._assign_colors(c)

            if not any(t.legendgroup == condition for t in self.fig.data):
                self.add_line(y=[None],name=f"{condition}", color=line_color,legendgroup=condition, showlegend=True )

            for i, channel in enumerate(data[condition]):
                average = data[condition][channel]["average"]
                standard_dev = data[condition][channel]["standard_dev"]

                # populate the figure
                row = i + 1
                col = self.conditions.index(condition) + 1
                self.add_line(y=average, row=row, col=col, name=f"{condition} - {channel}", color=line_color, legendgroup=condition, showlegend=False ) # color='#1F77B4')
                self.add_errorbar(y=average, yerr=standard_dev, row=row, col=col, color = shade_color,) #="rgba(31,119,180,0.3)")

        self.show()

    def _calculate_average(self):
        """Calculates the average timeseries for the channels"""
        # Initialize dictionary to store data

        data_new = {c: {ch: [] for ch in self.channels} for c in self.conditions}

        for fl in self.zoo_files:
            data = zload(fl)
            condition = self._get_condition_from_path(fl)

            # Create dataframe from the two conditions.
            for channel in self.channels:
                try:
                    ch_data_line = data[channel]["line"]
                    data_new[condition][channel].append(ch_data_line)
                except KeyError:
                    print(f"Channel {channel} not found in file {fl}")

        # Average per condition per channel
        average_dict = {c: {ch: {} for ch in self.channels} for c in self.conditions}
        for c, condition in enumerate(data_new):
            for i, channel in enumerate(data_new[condition]):
                line_data = data_new[condition][channel]
                array_data = np.array(line_data)
                average = np.nanmean(array_data, axis=0)
                standard_dev = np.nanstd(array_data, axis=0)

                average_dict[condition][channel].update({"average": average, "standard_dev": standard_dev})

        return average_dict

    def add_line(self, y, x=None, row=1, col=1, name=None, color=None, legendgroup=None, showlegend=True, customdata=None, hovertemplate=None):
        trace = go.Scatter(x=x, y=y, mode="lines", name=name,
                           line=dict(color=color), showlegend=showlegend, legendgroup=legendgroup,
                           customdata=customdata, hovertemplate=hovertemplate)
        self.fig.add_trace(trace, row=row, col=col)

    def add_marker(self, y, x, row=1, col=1, name=None, color=None, legendgroup=None, showlegend=True, customdata=None, hovertemplate=None):
        trace = go.Scatter(x=x, y=y, mode="markers", name=name,
                           line=dict(color=color), showlegend=showlegend, legendgroup=legendgroup,
                           customdata=customdata, hovertemplate=hovertemplate)
        self.fig.add_trace(trace, row=row, col=col)

    def add_errorbar(self, y, yerr, row=1, col=1, color=None):
        upper_bound = y + yerr
        lower_bound = y - yerr

        trace_lower = go.Scatter(y=lower_bound,
                                 line=dict(color='rgba(0,0,0,0)'),
                                 showlegend=False,
                                 )

        trace_upper = go.Scatter(y=upper_bound,
                           fill="tonexty",
                           fillcolor=color,
                           line=dict(color='rgba(0,0,0,0)'),
                           showlegend=False)

        self.fig.add_trace(trace_lower, row=row, col=col)
        self.fig.add_trace(trace_upper, row=row, col=col)

    def show(self, title=None):
        # Dynamic sizing
        base_height = 350
        base_width = 550
        height = base_height * self.rows
        width = base_width * self.cols

        # Default title if not provided
        if title is None:
            if self.cols == 1:
                title = "Combined Conditions"
            else:
                title = "Conditions by Channel"

        # Update layout
        self.fig.update_layout(
            height=height,
            width=width,
            title=dict(text=title, x=0.5, font=dict(size=24)),
            template="simple_white",
            margin=dict(l=50, r=50, t=50, b=50),
            font=dict(size=18), showlegend=True,
        )

        self.fig.show()

    def save(self, file_name, extension="html", folder=None):
        if folder is None:
            folder = self.fld

        os.makedirs(folder, exist_ok=True)
        if extension == "html":
            self.fig.write_html(os.path.join(folder, f"{file_name}.{extension}"))
        else:
            self.fig.write_image(os.path.join(folder, f"{file_name}.{extension}"))



class EnsemblerQualityChecker:
    def __init__(self, figure, out_folder):
        self.figure = figure
        self.out_folder = out_folder
        self.app = Dash(__name__)
        self._built_layout()
        self._register_callbacks()


    def _built_layout(self):
        self.app.layout = html.Div([
            # The graph
            html.Div([
                dcc.Graph(id="ensemble-graph", figure=self.figure, clear_on_unhover=True),
            ]),
            html.Hr(),
            # click output
            html.Div([
                html.H4("Last click"),
                html.Pre(id="last-click", style={"whiteSpace": "pre-wrap"}),
                html.H4("Clicks captured"),
                html.Pre(id="click-count")
            ]),
            # Download button
            html.Div([
                html.Button("Download CSV", id="btn-download", n_clicks=0),
                dcc.Download(id="download-csv"),
                dcc.Store(id="click-store", data=[])
            ]),
        ])

    def _register_callbacks(self):
        app=self.app

        @app.callback(
            Output("last-click", "children"),
            Output("click-count", "children"),
            Output("click-store", "data"),
            Output("ensemble-graph", "figure"),
            Input("ensemble-graph", "clickData"),
            State("click-store", "data"),
            State("ensemble-graph", "figure"),
            prevent_initial_call=True
        )
        def save_and_remove(clickData, clicks, fig):
            if not clickData or fig is None:
                return no_update, no_update, clicks, no_update

            pt = clickData["points"][0]
            # Ignore helper/legend traces that use y=[None]
            if pt.get("y") is None or pt.get("curveNumber") is None:
                return no_update, no_update, clicks, no_update

            # Build record (flat customdata: [subject, channel, condition, file, row, col, index, value])
            cd = pt.get("customdata") or []
            record = {
                "subject": cd.get("subject"),
                "channel": cd.get("channel"),
                "condition": cd.get("condition"),
                "source_file": cd.get("source_file"),
                "row": cd.get("row"),
                "col": cd.get("col"),
                "index": cd.get("index"),
                "value": cd.get("value"),
                # native plotly info as well
                "curveNumber": pt.get("curveNumber"),
                "pointNumber": pt.get("pointNumber"),
                "x": pt.get("x"),
                "y": pt.get("y"),
            }

            # Append & persist
            clicks = (clicks or []) + [record]
            try:
                out_dir = os.path.join(self.out_folder, "click_exports")
                os.makedirs(out_dir, exist_ok=True)
                # pd.DataFrame(clicks).to_csv(os.path.join(out_dir, "clicks_latest.csv"), index=False)
            except Exception:
                pass  # keep UI responsive even if write fails

            # Remove the clicked trace
            data = list(fig.get("data", []))
            idx = pt["curveNumber"]
            if 0 <= idx < len(data):
                t = data[idx]
                if t.get("type") == "scatter" and t.get("mode") in ("lines", "lines+markers", "markers"):
                    data.pop(idx)
                    fig["data"] = data
                    fig.setdefault("layout", {})["uirevision"] = "ensembler"  # preserve zoom/state

            return json.dumps(record, indent=2), f"Total clicks: {len(clicks)}", clicks, fig

        @app.callback(
            Output("download-csv", "data"),
            Input("btn-download", "n_clicks"),
            State("click-store", "data"),
            prevent_initial_call=True
        )
        def download_csv(n, clicks):
            if not clicks:
                return no_update
            df = pd.DataFrame(clicks)
            # For client-side download
            return dcc.send_data_frame(df.to_csv, "ensembler_clicks.csv", index=False)

    def run(self, **kwargs):
        # Default values if not provided
        kwargs.setdefault("host", "127.0.0.1")
        kwargs.setdefault("port", 8050)
        kwargs.setdefault("debug", True)
        self.app.run(**kwargs)



import os
import json
import pandas as pd
from dash import Dash,dcc, html, Input, Output, State, no_update, callback_context

from biomechzoo.visualization.ensembler import Ensembler

def run_quality_check(fld, ch, out_folder, subj_pattern, conditions=None, name_contains=None, event_name=None):

    if isinstance(ch, str):
        ch = [ch]

    if isinstance(conditions, str):
        conditions = [conditions]

    if isinstance(name_contains, str):
        name_contains = [name_contains]

    ensembler = Ensembler(fld=fld, ch=ch, conditions=conditions, name_contains=name_contains, subj_pattern=subj_pattern)
    ensembler.quality_check_cycles(event_name=event_name)

    # snapshot of base figure
    base_fig = ensembler.fig.to_dict() if hasattr(ensembler.fig, "to_dict") else {
        "data": ensembler.fig.data,
        "layout": ensembler.fig.layout,
    }
    # base_fig.setdefault("layout", {})["uirevision"] = "ensembler"

    external_stylesheets = [
        {"href": ("https://fonts.googleapis.com/css2?"
                  "family=Lato:wght@400;700&display=swap"),
         "rel": "stylesheet", }
    ]

    app = Dash(__name__, external_stylesheets=external_stylesheets)

    app.layout = html.Div(children=[
        # ---- Header----
        html.Div(
            children=[
                html.H2(children="Quality check", className="header-title"),
                html.P(children='''
                Click on the lines that are wrongfully segmented to remove, 
                and press download to save the csv file for removed cycles
                ''', className="header-description"),
                    ],className="header",
                ),
        # ----Layout of the figure----
        html.Div(
            children=
                dcc.Graph(id='cycle-graph',
                    figure={"data": ensembler.fig.data,
                            "layout":{"title": {"text": f"{ch[0]}", "x": 0.05, "xanchor": "left"},
                                      }
                            },
                          ), className="card",
            style={'width': '60%', 'float': 'right', 'display': 'inline-block',},

        ),

        # ----Controls undo/reset----
        html.Div(
            children=[
                html.Button("Undo last removal", id="btn-undo", n_clicks=0, disabled=True),
                html.Button("Reset figure", id="btn-reset", n_clicks=0),
                dcc.Store(id="undo-stack", data=[]),
                dcc.Store(id="base-figure", data=base_fig),
            ], className="card",
        ),

        # ----Click functionality----
        html.Div(children=[
            html.H4("Last click"),
            html.Pre(id="last-click"),
            html.H4("Clicks captured"),
            html.Pre(id="click-count"),
        ]),
        # ----Download button----
        html.Div([
            html.Button("Download CSV", id="btn-download", n_clicks=0),
            dcc.Download(id="download-csv"),
            dcc.Store(id="click-store", data=[])

        ]),
        html.Div([
            html.Img(src=app.get_asset_url("(Preferred) - Red on white logo (1).png"), alt="The official McGill logo",
                     style={"width": "15%", 'float': 'right'}),
        ], style={'display': 'inline-block', 'vertical-align': 'bottom'}),
        ], className="wrapper"
    )

    @app.callback(
        Output("last-click", "children"),
        Output("click-count", "children"),
        Output("click-store", "data"),
        Output("cycle-graph", "figure"),
        Output("undo-stack", "data"),
        Output("btn-undo", "disabled"),
        Input("cycle-graph", "clickData"),
        Input("btn-undo", "n_clicks"),
        Input("btn-reset", "n_clicks"),
        State("click-store", "data"),
        State("cycle-graph", "figure"),
        State("undo-stack", "data"),
        State("base-figure", "data"),
        prevent_initial_call=True
    )
    def multi_action(clickData, n_undo, n_reset, clicks, fig, undo_stack, base_fig):
        # Default passthroughs
        last_click_msg = no_update
        click_count_msg = no_update
        clicks_out = clicks
        fig_out = fig
        undo_out = undo_stack or []
        undo_disabled_out = len(undo_out) == 0

        # Which input fired?
        ctx = callback_context
        trigger = ctx.triggered[0]["prop_id"] if ctx.triggered else ""

        # ---- RESET ----
        if trigger.startswith("btn-reset.n_clicks"):
            if base_fig:
                fig_out = {
                    "data": base_fig.get("data", []),
                    "layout": base_fig.get("layout", {})
                }
                # fig_out.setdefault("layout", {})["uirevision"] = "ensembler"
                undo_out = []
                undo_disabled_out = True
            return last_click_msg, click_count_msg, clicks_out, fig_out, undo_out, undo_disabled_out

        # ---- UNDO ----
        if trigger.startswith("btn-undo.n_clicks"):
            if undo_out:
                last = undo_out[-1]
                idx = last.get("index")
                trace = last.get("trace")
                click_idx = last.get("click_idx")

                data = list(fig_out.get("data", []))
                insert_at = max(0, min(idx if isinstance(idx, int) else len(data), len(data)))
                data.insert(insert_at, trace)
                fig_out["data"] = data
                fig_out.setdefault("layout", {})["uirevision"] = "ensembler"

                if isinstance(click_idx, int) and 0 <= click_idx < len(clicks_out or []):
                    clicks_out = list(clicks_out)
                    # Choose ONE label and use it consistently:
                    clicks_out[click_idx]["action"] = "reinstated"  # or "restated" if you prefer

                undo_out = undo_out[:-1]

            undo_disabled_out = len(undo_out) == 0
            return last_click_msg, click_count_msg, clicks_out, fig_out, undo_out, undo_disabled_out

        # ---- CLICK-TO-REMOVE ----
        if trigger.startswith("cycle-graph.clickData"):
            if not clickData or fig_out is None:
                return last_click_msg, click_count_msg, clicks_out, no_update, undo_out, undo_disabled_out

            pt = clickData["points"][0]
            if pt.get("y") is None or pt.get("curveNumber") is None:
                return last_click_msg, click_count_msg, clicks_out, no_update, undo_out, undo_disabled_out

            cd = pt.get("customdata") or {}

            def _get(k, default=None):
                if isinstance(cd, dict):
                    return cd.get(k, default)
                if isinstance(cd, list) and cd and isinstance(cd[0], dict):
                    return cd[0].get(k, default)
                return default

            record = {
                "subject": _get("subject"),
                "channel": _get("channel"),
                "condition": _get("condition"),
                "source_file": _get("source_file"),
                "row": _get("row"),
                "col": _get("col"),
                "index": _get("index"),
                "value": _get("value"),
                "curveNumber": pt.get("curveNumber"),
                "pointNumber": pt.get("pointNumber"),
                "x": pt.get("x"),
                "y": pt.get("y"),
                "action": "remove"
            }

            clicks_out = (clicks_out or []) + [record]
            click_idx = len(clicks_out) - 1  # index of the removal record in click-store

            try:
                out_dir = os.path.join(out_folder, "click_exports")
                os.makedirs(out_dir, exist_ok=True)
                # pd.DataFrame(clicks_out).to_csv(os.path.join(out_dir, "clicks_latest.csv"), index=False)
            except Exception:
                pass

            data = list(fig_out.get("data", []))
            idx = pt["curveNumber"]
            if 0 <= idx < len(data):
                removed_trace = data.pop(idx)
                fig_out["data"] = data
                fig_out.setdefault("layout", {})["uirevision"] = "ensembler"
                undo_out = (undo_out or []) + [{"index": idx, "trace": removed_trace, "click_idx": click_idx}]
                undo_disabled_out = False

            last_click_msg = json.dumps(record, indent=2)
            click_count_msg = f"Total clicks: {len(clicks_out)}"

            return last_click_msg, click_count_msg, clicks_out, fig_out, undo_out, undo_disabled_out

        # Fallback: nothing recognized
        return last_click_msg, click_count_msg, clicks_out, fig_out, undo_out, undo_disabled_out

    # def save_and_remove(clickData, clicks, fig, undo_stack):
    #     if not clickData or fig is None:
    #         return no_update, no_update, clicks, no_update, no_update, no_update
    #
    #     pt = clickData["points"][0]
    #     # Ignore helper/legend traces that use y=[None]
    #     if pt.get("y") is None or pt.get("curveNumber") is None:
    #         return no_update, no_update, clicks, no_update, no_update, no_update
    #
    #     # Build record (flat customdata: [subject, channel, condition, file, row, col, index, value])
    #     cd = pt.get("customdata") or []
    #     record = {
    #         "subject": cd.get("subject"),
    #         "channel": cd.get("channel"),
    #         "condition": cd.get("condition"),
    #         "source_file": cd.get("source_file"),
    #         "row": cd.get("row"),
    #         "col": cd.get("col"),
    #         "index": cd.get("index"),
    #         "value": cd.get("value"),
    #         # native plotly info as well
    #         "curveNumber": pt.get("curveNumber"),
    #         "pointNumber": pt.get("pointNumber"),
    #         "x": pt.get("x"),
    #         "y": pt.get("y"),
    #     }
    #
    #     # Append & persist
    #     clicks = (clicks or []) + [record]
    #     try:
    #         out_dir = os.path.join(out_folder, "click_exports")
    #         os.makedirs(out_dir, exist_ok=True)
    #         # pd.DataFrame(clicks).to_csv(os.path.join(out_dir, "clicks_latest.csv"), index=False)
    #     except Exception:
    #         pass  # keep UI responsive even if write fails
    #
    #     # Remove the clicked trace and push to undo stack
    #     data = list(fig.get("data", []))
    #     idx = pt["curveNumber"]
    #     if 0 <= idx < len(data):
    #         removed_trace = data[idx]
    #         fig["data"] = data
    #         fig.setdefault("layout", {})["uirevision"] = "ensembler"
    #
    #         undo_stack = (undo_stack or []) + [{
    #             "index": idx,
    #             "trace": removed_trace
    #         }]
    #
    #     # enable undo if undo-stack is not empty
    #     undo_disabled = len(undo_stack or []) == 0
    #
    #
    #     return json.dumps(record, indent=2), f"Total clicks: {len(clicks)}", clicks, fig, undo_stack


    @app.callback(
        Output("download-csv", "data"),
        Input("btn-download", "n_clicks"),
        State("click-store", "data"),
        prevent_initial_call=True
    )
    def download_csv(n, clicks_out):
        if not n or not clicks_out:
            return no_update

        removed = [
            r for r in (clicks_out or [])
            if r.get("action") == "remove" and not r.get("undone", False)
        ]

        if not removed:
            return no_update
        df = pd.DataFrame(removed)
        # For client-side download
        return dcc.send_data_frame(df.to_csv, "removed_cycles.csv", index=False)

    app.run(debug=True)







import numpy as np
import scipy.signal as signal

def movement_onset(yd, fsamp, constants):
    """
    Extracts movement onset based on the average and standard deviation of a sliding window
    Standard thresholds for running are mean_thresh=1.2, std_thresh=0.2. For walking mean_thresh=0.6, std_thresh=0.2.

    yd: 1d array of the vector
    fsamp: sampling frequency
    constants: [mean_thresh, std_thresh]
    etype: 'movement_onset' or 'movement_offset'
    """
    acc_mag = yd.copy()
    acc_mag_filtered = bw_filter(data=acc_mag, fsamp=fsamp, N=4, fc=20, btype="low")
    features, timestamps = sliding_window_features(ch_data=acc_mag_filtered, fsamp=fsamp)

    mean_thresh, std_thresh = constants[0], constants[1]
    min_thresh = 0.1
    onset_time = None
    while onset_time is None and mean_thresh > min_thresh:
        # ----Check if already moving----
        if check_already_moving(features=features, mean_thresh=mean_thresh, std_thresh=std_thresh):
            onset_time = timestamps[0]

        # ----Try detecting onset----
        else:
            onset_index = detect_movement_onset(features, fsamp, mean_thresh, std_thresh)
            if onset_index is not None:
                onset_time = timestamps[onset_index]
            else:
                # relax thresholds for next iteration
                mean_thresh /=2
                std_thresh /= 2

    return onset_time


def movement_offset(yd, fsamp, constants):

    # ----extract the constants----
    mean_thresh, std_thresh = constants[0], constants[1]
    min_thresh = 0.1
    onset_time = None

    # ----Reverse, filter and extract features from the signa;----
    acc_mag = yd.copy()
    acc_mag = acc_mag[::-1]

    acc_mag_filtered = bw_filter(data=acc_mag, N=4, fc=20, btype="low", fsamp=fsamp)
    features, timestamps = sliding_window_features(ch_data=acc_mag_filtered, fsamp=fsamp)

    # ----reverse timestamps----
    timestamps = timestamps[::-1]

    while onset_time is None and mean_thresh > min_thresh:
        # ----Check if already moving----
        if check_already_moving(features=features, mean_thresh=mean_thresh, std_thresh=std_thresh):
            onset_time = timestamps[0]

        # ----Try detecting onset----
        else:
            onset_index = detect_movement_onset(features, fsamp, mean_thresh, std_thresh)
            if onset_index is not None:
                onset_time = timestamps[onset_index]
            else:
                # relax thresholds for next iteration
                mean_thresh /= 2
                std_thresh /= 2

    return onset_time


def sliding_window_features(ch_data, fsamp):
    # ----sliding window features----
    features = []
    timestamps = []
    window_size = 2 * fsamp  # windows van 2 seconds
    step_size = 1 * fsamp  # with an overlap of 1 seconds

    for start in range(0, len(ch_data) - window_size, step_size):
        segment = ch_data[start:start + window_size]
        mean_val = segment.mean()
        std_val = segment.std()
        # entropy = -np.sum((segment / np.sum(segment)) * np.log2(segment / np.sum(segment) + 1e-12))
        timestamps.append(start)
        features.append((mean_val, std_val))

    return np.array(features), np.array(timestamps)


def check_already_moving(features, mean_thresh=1.2, std_thresh=0.2):
    initial_window = features[:5]  # First few seconds
    if np.all(initial_window[:, 0] > mean_thresh) and np.all(initial_window[:, 1] > std_thresh):
        return True
    return False

def detect_movement_onset(features, fs, mean_thresh=1.2, std_thresh=0.2, min_duration=3):
    movement_flags = (features[:, 0] > mean_thresh) & (features[:, 1] > std_thresh)
    onset_index = None
    for i in range(len(movement_flags) - int(min_duration * fs / 50)):
        if np.all(movement_flags[i:i + int(min_duration * fs / 50)]):
            onset_index = i
            break
    return onset_index if onset_index is not None else None

def bw_filter(data, fsamp, N, fc, btype, output="ba"):
    """
    Basic zero-phase butterworth filter.
    :param data: 1xN array
    :param N: filter order
    :param fc: cutoff frequency
    :param btype: filter type
    :return: filtered data 1xN.

    Args:
        output:
    """

    answer = fc
    Fn = fsamp / 2
    Fnrad = 2 * np.pi * Fn
    Fc = 2 * np.pi * answer
    Wn = [Fc / Fnrad]
    # [b, a] = butter(4, Wn, 'low');
    # local_breast = -(filtfilt(b, a, local_breast_raw(:,:)));

    [b, a] = signal.butter(N, Wn=Wn, btype=btype, output=output)
    filtered_data = signal.filtfilt(b=b, a=a, x=data)

    return filtered_data


import numpy as np
from scipy.interpolate import interp1d


def normalize_line(channel_data, nlength=101):
    """
    Channel-level: interpolate channel data to target length.
    Assumes channel_data is a 1D or 2D numpy array.
    """
    original_length = channel_data.shape[0]

    if original_length == nlength:
        return channel_data

    x_original = np.linspace(0, 1, original_length)
    x_target = np.linspace(0, 1, nlength)

    if channel_data.ndim == 1:
        f = interp1d(x_original, channel_data, kind='linear')
        channel_data_norm = f(x_target)
    else:
        channel_data_norm = np.zeros((nlength, channel_data.shape[1]))
        for i in range(channel_data.shape[1]):
            f = interp1d(x_original, channel_data[:, i], kind='linear')
            channel_data_norm[:, i] = f(x_target)

    return channel_data_norm




if __name__ == '__main__':
    # --- 1D TESTS ---
    data_1d = np.array([0, 1, 2, 3, 4])
    print("Original 1D shape:", data_1d.shape)

    # Case 1: same length
    same = normalize_line(data_1d, nlength=5)
    print("Same length test passed:", np.allclose(same, data_1d))

    # Case 2: upsample
    upsampled = normalize_line(data_1d, nlength=10)
    print("Upsampled 1D shape:", upsampled.shape)
    print("Upsampled 1D first/last values:", upsampled[0], upsampled[-1])

    # Case 3: downsample
    downsampled = normalize_line(data_1d, nlength=3)
    print("Downsampled 1D shape:", downsampled.shape)
    print("Downsampled 1D first/last values:", downsampled[0], downsampled[-1])

    print("\nAll tests completed.")


import numpy as np
from scipy.signal import find_peaks, butter, filtfilt

def imu_mcgrath(ch_line, fsamp, min_stance_t, is_filtered=False):
    """
    This function detects the steps based on the method of McGrath et al. (2012) https://doi.org/10.1007/s12283-012-0093-8
    in short, the first minimum after a local maximum is the heel strike. The local maxima are the mid-swing.
    Data should be filtered
    """

    if is_filtered:
        yd = ch_line
    else:
        # Butterworth filter
        order = 5
        Fc = 5
        Wn = Fc / (fsamp / 2)
        [b, a] = butter(order, Wn, btype='low')
        yd = filtfilt(b, a, ch_line)


    # Identify midswing peaks
    t1 = round(fsamp / 2)
    potential_midswing_ind, _ = find_peaks(yd, distance=t1)
    potential_midswing_mag = yd[potential_midswing_ind]

    # Thresholds for midswing
    th2 = 0.8 * np.mean(yd[yd > np.mean(yd)])
    mask = potential_midswing_mag >= th2
    potential_midswing_ind = potential_midswing_ind[mask]
    potential_midswing_mag = potential_midswing_mag[mask]

    th1 = 0.3 * potential_midswing_mag

    # Find minima
    minima_ind, _ = find_peaks(-yd)
    minima_mag = -yd[minima_ind]

    # Validate midswing peaks
    valid_inds = []
    for i in range(len(potential_midswing_ind) - 1, -1, -1):
        peak_idx = potential_midswing_ind[i]
        preceding_min = minima_ind[minima_ind < peak_idx]
        if preceding_min.size > 0:
            closest_min_idx = preceding_min[-1]
            pos = np.where(minima_ind == closest_min_idx)[0][0]
            if (potential_midswing_mag[i] - minima_mag[pos]) >= th1[i]:
                valid_inds.append(peak_idx)
    potential_midswing_ind = np.array(valid_inds)

    # Additional thresholds
    th3 = 0.8 * abs(np.mean(yd[yd < np.mean(yd)]))
    th4 = 0.8 * np.mean(yd[yd < np.mean(yd)])
    th5 = np.mean(yd)
    th6 = 2 * th3

    maxima_ind, _ = find_peaks(yd)
    maxima_mag = yd[maxima_ind]

    IC, TC = [], []
    t2 = round(1.5 * fsamp)

    # Loop through confirmed midswing peaks
    for step_idx in range(len(potential_midswing_ind) - 1, -1, -1):
        peak_idx = potential_midswing_ind[step_idx]

        # IC candidates (minima after midswing)
        end_idx = min(peak_idx + t2, len(yd))
        ic_candidates, _ = find_peaks(-yd[peak_idx:end_idx])
        ic_candidates = ic_candidates + peak_idx
        ic_mags = yd[ic_candidates]

        # Filter IC by threshold
        ic_candidates = ic_candidates[ic_mags < th5]
        ic_mags = ic_mags[ic_mags < th5]

        # Validate IC with preceding maxima
        for ic_idx in ic_candidates:
            preceding_max = maxima_ind[maxima_ind < ic_idx]
            if preceding_max.size > 0:
                closest_max_idx = preceding_max[-1]
                if yd[closest_max_idx] >= yd[ic_idx] + th3:
                    IC.append(ic_idx)
                    break

        # TC candidates (minima before midswing)
        start_idx = max(peak_idx - t2, 0)
        tc_candidates, _ = find_peaks(-yd[start_idx:peak_idx])
        tc_candidates = tc_candidates + start_idx
        tc_mags = yd[tc_candidates]

        # Filter TC by threshold
        tc_candidates = tc_candidates[tc_mags < th4]
        tc_mags = tc_mags[tc_mags < th4]

        # Validate TC with following maxima
        for tc_idx in tc_candidates:
            following_max = maxima_ind[maxima_ind > tc_idx]
            if following_max.size > 0:
                closest_max_idx = following_max[0]
                if yd[closest_max_idx] >= yd[tc_idx] + th6:
                    TC.append(tc_idx)
                    break

    # Crash handling
    IC, TC = crash_catch(int(min_stance_t * fsamp / 1000), IC, TC)

    return IC, TC

def crash_catch(min_stance_samples, IC, TC):
    # Ensure IC and TC arrays are same length and valid
    IC = np.array(IC)
    TC = np.array(TC)
    if len(IC) != len(TC):
        min_len = min(len(IC), len(TC))
        IC = IC[:min_len]
        TC = TC[:min_len]
    return IC, TC


def update_channel_list(data, section='Video', ch_add=None, ch_remove=None):
    """
    Updates the channel list of a section by adding or removing channels.

    Args:
        data (dict): Zoo data dictionary.
        section (str): Section name ('Video', 'Analog', etc.).
        ch_add (str or list of str, optional): Channel(s) to add.
        ch_remove (str or list of str, optional): Channel(s) to remove.

    Returns:
        dict: The updated zoo dictionary (same object as input).
    """
    ch_list = data['zoosystem'][section]['Channels']

    # Normalize to list
    if isinstance(ch_add, str):
        ch_add = [ch_add]
    if isinstance(ch_remove, str):
        ch_remove = [ch_remove]

    # Add channels
    if ch_add is not None:
        for ch in ch_add:
            if ch not in ch_list:
                ch_list.append(ch)

    # Remove channels
    if ch_remove is not None:
        for ch in ch_remove:
            if ch in ch_list:
                ch_list.remove(ch)

    data['zoosystem'][section]['Channels'] = ch_list
    return data


def findfield(data, target_event):
    """ searches in zoo data for the event value and channel name associated with target_event

    Arguments:
        data. dict. zoo data
        target_event, str. Name of event to search for
    Returns:
        events, list of len 2. First is index (exd) and second is value (eyd)
        channel, str. Name of channel associated with event
    """
    for channel, content in data.items():
        if channel == 'zoosystem':
            continue
        events = content.get('event', {})
        if target_event in events:
            val = events[target_event]
            return val, channel

    return None, None





import os
import pandas as pd
from biomechzoo.utils.engine import engine
from biomechzoo.utils.zload import zload
from biomechzoo.utils.findfield import findfield  # assuming this exists

def eventval(fld, dim1=None, dim2=None, ch=None, localevts=None, globalevts=None, anthroevts=None):
    """
    Extract event values from .zoo files and compile into a pandas DataFrame.

    Parameters
    ----------
    fld : str
        Path to the root data folder containing .zoo files.
    dim1 : list of str, optional
        List of conditions (subfolder names under fld).
    dim2 : list of str, optional
        List of participant identifiers.
    ch : list of str
        List of channels to extract events from.
    localevts : list of str, optional
        List of local events.
    globalevts : list of str, optional
        List of global events.
    anthroevts : list of str, optional
        List of events stored in the metadata.Usually anthropometric data

    Returns
    -------
    pd.DataFrame
        Columns: ['condition', 'subject', 'file', 'event_name', 'event_value']
    """

    zoo_files = engine(fld, extension='.zoo')
    results = []
    files_excluded = []
    for fl in zoo_files:
        data = zload(fl)
        fname = os.path.basename(fl)
        condition = next((c for c in (dim1 or []) if c in fl), '')
        subject = next((s for s in (dim2 or []) if s in fl), '')

        # Skip file if condition or subject is empty

        if not condition or not subject:
            files_excluded.append(fl)
            continue
        else:
            print('processing {}'.format(fl))

        # --- Local events ---
        if localevts and ch:
            print('extracting local events...')
            for channel in ch:
                if channel not in data:
                    print('channel {} not found'.format(channel))
                    continue
                for evt in localevts:
                    try:
                        exd = int(data[channel]['event'][evt][0])  # xdata
                        eyd = data[channel]['event'][evt][1]  # ydata
                        results.append({
                            'condition': condition,
                            'subject': subject,
                            'file': fname,
                            'event_name': f"{channel}_{evt}",
                            'event_index': exd,
                            'event_value': eyd
                        })
                    except (KeyError, IndexError, TypeError) as e:
                        print(f"Local event '{evt}' not found in channel '{channel}' for file '{fname}'")

        # --- Global events ---
        if globalevts and ch:
            print('extracting global events...')
            for evt in globalevts:
                # use findfield to locate where the global event is stored
                evt_val, evt_ch = findfield(data, evt)
                if not evt_ch:
                    print(f"Global event '{evt}' not found in any channel for file '{fname}'")
                    continue

                exd = int(evt_val[0])
                for channel in ch:
                    if channel not in data:
                        print(f"Skipping {evt}: channel {channel} not in data")
                        continue
                    try:
                        eyd = data[channel]['line'][exd]
                        results.append({
                            'condition': condition,
                            'subject': subject,
                            'file': fname,
                            'event_name': f"{channel}_{evt}",
                            'event_index': exd,
                            'event_value': eyd
                        })
                    except (IndexError, KeyError, TypeError) as e:
                        print(f"Global event '{evt}' index out of range in channel '{channel}' for file '{fname}': {e}")

        if anthroevts and ch:
            raise NotImplementedError
            print('extracting anthropometric events...')
            for evt in globalevts:
                # use findfield to locate where the global event is stored
                evt_val, _ = findfield(data, evt)

                if evt_val:
                    results.append({
                        'condition': condition,
                        'subject': subject,
                        'file': fname,
                        'event_name': evt,
                        'event_index': evt_val[0],
                        'event_value': evt_val[1]
                    })

    return pd.DataFrame(results)



def common_substring_join(strings):
    # Split each string into parts
    split_lists = [s.split('_') for s in strings]

    # Transpose so we compare column-wise
    common_parts = []
    for parts in zip(*split_lists):
        if len(set(parts)) == 1:   # All strings share this part
            common_parts.append(parts[0])
        else:
            # As soon as one position differs, skip but keep checking further ones
            continue

    return "_".join(common_parts)

if __name__ == '__main__':
    # Example
    a = [
        'a_pelvis_antpost_tilt_corr',
        'a_pelvis_ml_tilt_corr',
        'a_pelvis_vert_tilt_corr'
    ]

    print(common_substring_join(a))


import copy
import numpy as np
from biomechzoo.utils.update_channel_list import update_channel_list

def explodechannel_data(data, channels=None):
    """ Explodes 3D channels (n x 3 arrays) into separate X, Y, Z channels.

    Arguments:
        data (dict): Zoo data loaded from a file
        channels (list of str or None): Channels to explode.
            If None, explode all channels with 'line' shaped (n x 3).

    Returns:
        data_new (dict): Modified zoo dictionary with exploded channels.
    """
    data_new = copy.deepcopy(data)

    # Find default channels if none provided
    if channels is None:
        channels = []
        for ch in data_new:
            if ch == 'zoosystem':
                continue
            ch_data = data_new[ch]['line']
            if ch_data.ndim == 2 and ch_data.shape[1] == 3:
                channels.append(ch)

    # Explode each channel
    for ch in channels:
        if ch not in data_new:
            print('Warning: channel {} not found, skipping.'.format(ch))
            continue

        # find section
        if ch in data_new['zoosystem']['Video']['Channels']:
            section = 'Video'
        elif ch in data_new['zoosystem']['Analog']['Channels']:
            section = 'Analog'
        else:
            raise ValueError('Unknown section for channel {}'.format(ch))

        ch_data = data_new[ch]['line']
        if ch_data.ndim != 2 or ch_data.shape[1] != 3:
            print(f"Warning: channel '{ch}' 'line' is not n x 3 shape, skipping.")
            continue

        # Add exploded channels
        x, y, z = ch_data[:, 0], ch_data[:, 1], ch_data[:, 2]
        for axis, line in zip(['_x', '_y', '_z'], [x, y, z]):
            key = ch + axis

            # Bring the old events to _x only
            if axis == '_x':
                edata = data_new[ch]['event']
            else:
                edata = {}

            data_new[key] = {
                'line': line,
                'event': edata
            }

            # Update channel list and assign back
            data_new = update_channel_list(data_new, section=section, ch_add=key)

        # Remove original channel from list and dict
        data_new = update_channel_list(data_new, section=section, ch_remove=ch)
        data_new.pop(ch)


    return data_new


if __name__ == '__main__':
    # -------TESTING--------
    import os
    from biomechzoo.utils.zload import zload
    current_dir = os.path.dirname(os.path.abspath(__file__))
    project_root = os.path.dirname(os.path.dirname(os.path.dirname(current_dir)))
    fl = os.path.join(project_root, 'data', 'other', 'HC030A05.zoo')

    # load  zoo file
    data = zload(fl)
    ch_old_name = 'RKneeAngles'
    ch_new_name = 'RightKneeAngles'
    data = explodechannel_data(data)




from scipy.spatial.transform import Rotation as R
import numpy as np

def load_quats(data:dict, prefix:str) -> np.ndarray:
    """
    Returns a stacked np.ndarray containing the w, x, y, z components of a quaternion in scalar first order.

    Note:           the function assumes that data will have a prefix before quaternions from different segments.
                    For example:

                    data.keys() = [LSh_Quat_W, LSh_Quat_X, ... LF_Quat_W, LF_Quat_X, ...]

                    load_quats(data, prefix='LF_') -> returns LF_Quat_W, LF_Quat_X, LF_Quat_Y, LF_Quat_Z

    :param data:    dict containing the sensor data
    :param prefix:  the prefix defining the segment that is being loaded
    :return:        stacked np.ndarray containing the w, x, y, z components of the sensor from the desired sensor
    """

    # Define the keys to search for segment data
    base = ["W", "X", "Y", "Z"]
    keys = [f"{prefix}_Quat_{b}" for b in base]

    # Extract keys
    quat_components = [data[k]['line'] for k in keys]

    return np.column_stack(quat_components)

def imu_angles_data(data:dict, prox_prefix:str, dist_prefix:str, order:str) -> dict:
    """
    Compute Euler angles describing the orientation of the distal segment with respect to the proximal segment.

    :param data:            dict containing the quaternions for each sensor
    :param prox_prefix:     prefix defining the proximal segment
    :param dist_prefix:     prefix defining the distal segment
    :param order:           order of the IMU sensors. Note, the case of the order changes between intrinsic or extrinsic rotations.
                            For more information please reference the scipy.spatial.transform documentation:
                            https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.transform.Rotation.html

    :return:                dict containing the Euler angles with alpha, beta, and gamma representing the first, second,
                            and third rotations in the sequence, respectively. Results are in degrees.
    """

    # Load the quaternions from the proximal and distal segments
    q_prox = load_quats(data, prefix=prox_prefix)
    q_dist = load_quats(data, prefix=dist_prefix)

    # Convert to Rotation objects
    R_prox = R.from_quat(q_prox, scalar_first=True)
    R_dist = R.from_quat(q_dist, scalar_first=True)

    # Derive relative orientation
    R_rel = R_prox.inv() * R_dist

    # Convert to Euler angles using defined rotation order
    euler = R_rel.as_euler(order, degrees=True)

    angles = {
        f"{prox_prefix}_{dist_prefix}_alpha": {"line": euler[:, 0]},
        f"{prox_prefix}_{dist_prefix}_beta":  {"line": euler[:, 1]},
        f"{prox_prefix}_{dist_prefix}_gamma": {"line": euler[:, 2]},
    }

    data.update(angles)

    return data


import os

import pandas as pd


def combine_quats_to_csv(
    csv_files: list[str],
    prefixes: list[str],
    out_folder: str = None,
    out_filename: str = None
    ) -> str:

    """Concatenates time, quaternions, gyroscope, and accelerometer data
    from multiple CSV files into a single CSV file with prefixes defining segment."""

    if out_folder is None:
        out_folder = "combined_csvs"

    if out_filename is None:
        out_filename = "combined_sensors.csv"

    root = os.getcwd()
    save_folder = os.path.join(root, out_folder)
    os.makedirs(save_folder, exist_ok=True)

    time_col: str = "PacketCounter"
    quat_cols: list[str] = ["Quat_W", "Quat_X", "Quat_Y", "Quat_Z"]
    gyr_cols: list[str] = ["Gyr_X", "Gyr_Y", "Gyr_Z"]
    acc_cols: list[str] = ["Acc_X", "Acc_Y", "Acc_Z"]

    first_df = pd.read_csv(csv_files[0])
    time_df = first_df[[time_col]].rename(columns={time_col: "time"})

    all_sensor_dfs = []

    for csv_path, prefix in zip(csv_files, prefixes):
        df = pd.read_csv(csv_path)

        quat_df = df[quat_cols].rename(columns={c: f"{prefix}_{c}" for c in quat_cols})
        gyr_df  = df[gyr_cols].rename(columns={c: f"{prefix}_{c}" for c in gyr_cols})
        acc_df  = df[acc_cols].rename(columns={c: f"{prefix}_{c}" for c in acc_cols})

        all_sensor_dfs.extend([quat_df, gyr_df, acc_df])

    combined_df = pd.concat([time_df] + all_sensor_dfs, axis=1)

    out_file = os.path.join(save_folder, out_filename)
    combined_df.to_csv(out_file, index=False)
    print(f"Saved combined CSV to: {out_file}")

    return out_file



import numpy as np
from biomechzoo.processing.addchannel_data import addchannel_data

def compute_magnitude_line(x,y,z):
    magnitude = np.sqrt((x**2) + (y**2) + (z **2))
    return magnitude

def rectify_data(data, chs):
    """
    Take absolute value of channels
    """
    if type(chs) is str:
        chs = [chs]

    # extract channels from data
    for ch in chs:
        yd = data[ch]['line']
        yd_abs = rectify_line(yd)
        data = addchannel_data(data, ch_new_data=yd_abs, ch_new_name=ch + '_rectified')
    return data

def rectify_line(yd):
    return np.abs(yd)





import numpy as np
import math
import pandas as pd
from biomechzoo.processing.addchannel_data import addchannel_data

def tilt_algorithm_data(data,ch_vert, ch_medlat, ch_antpost, plot_or_not=None):

    # extract channels from data
    avert = data[ch_vert]['line']
    amedlat = data[ch_medlat]['line']
    aantpost = data[ch_antpost]['line']

    _, avert_corr, amedlat_corr, aantpost_corr = tilt_algorithm_line(avert, amedlat, aantpost)

    data = addchannel_data(data, ch_vert + '_tilt_corr', avert_corr)
    data = addchannel_data(data, ch_medlat + '_tilt_corr', amedlat_corr)
    data = addchannel_data(data, ch_antpost + '_tilt_corr', aantpost_corr)

    return data


def tilt_algorithm_line(avert, amedlat, aantpost):
    """
    TiltAlgorithm - to account for gravity and improper tilt alignment of a tri-axial trunk accelerometer.

    Step 1: Extract raw measured (mean) accelerations
    Step 2: Calculate tilt angles
    Step 3: Calculate horizontal dynamic accelerations vectors
    Step 4: Calculate estimated provisional vertical vector
    Step 5: Calculate vertical dynamic vector
    step 6.1:  Calculate the contribution of static components
    step 6.2 Transpose static component matrices
    step 7: Remove the static components from the templates of pre and post

    Parameters
    ----------
    avert : 1D-array
        data predominantly in vertical direction. Expressed in g's
    amedlat : 1D-array:
        data predominantly in medio-lateral direction. Expressed in g's
    aantpost : 1D-array
        data predominantly in anterior-posterior direction. Expressed in g's
    Returns
    -------
    df_corrected : Nx3 DataFrame
        the tilt corrected and gravity subtracted vertical, medio-lateral and anterior-posterior
        acceleration signals
    avert2: 1D-array
        the tilt corrected acceleration data in vertical direction
    amedlat2 : 1D-array
        the tilt corrected acceleration data in medio-lateral direction
    aantpost2: 1D-array
        the tilt corrected acceleration data in anterior-posterior direction
    Notes
    -----
    -  If average acceleration is above 5m/s^2, the signal will be corrected.

    """

    a_vt = avert.mean()
    a_ml = amedlat.mean()
    a_ap = aantpost.mean()

    # if average vertical acceleration is more than 5, data is expressed in m/s^2
    # Update signals to G's
    if a_vt > 5:
        avert /= 9.81
        amedlat /= 9.81
        aantpost /= 9.81

        a_vt = avert.mean()
        a_ml = amedlat.mean()
        a_ap = aantpost.mean()



    # if avert is negative than turn the sensor around.
    if a_vt < 0.5:
        avert *= -1
        amedlat *= -1
        a_vt = avert.mean()
        a_ml = amedlat.mean()

    # Anterior tilt
    TiltAngle_ap_rad = np.arcsin(a_ap)
    TiltAngle_ap_deg = math.degrees(TiltAngle_ap_rad)

    # mediolateral tilt
    TiltAngle_ml_rad = np.arcsin(a_ml)
    TiltAngle_ml_deg = math.degrees(TiltAngle_ml_rad)

    # Anterior posterior
    a_AP = (a_ap * np.cos(TiltAngle_ap_rad)) - (a_vt * np.sin(TiltAngle_ap_rad))
    # AMediolateral
    a_ML = (a_ml * np.cos(TiltAngle_ml_rad)) - (a_vt * np.sin(TiltAngle_ml_rad))

    # a_vt_prov = a_ap*Sin(theta_ap) + a_vt*Cos(theta_ap)
    a_vt_prov = (a_ap * np.sin(TiltAngle_ap_rad)) + (a_vt * np.cos(TiltAngle_ap_rad))

    # a_VT = a_ml*sin(theta_ml) + a_vt_prov*cos(theta_ml) - 1
    a_VT = (a_ml * np.sin(TiltAngle_ml_rad)) + (a_vt_prov * np.cos(TiltAngle_ml_rad)) - 1

    a_AP_static = a_ap - a_AP
    a_ML_static = a_ml - a_ML
    a_VT_static = a_vt - a_VT

    a_AP_static = np.transpose(a_AP_static)
    a_ML_static = np.transpose(a_ML_static)
    a_VT_static = np.transpose(a_VT_static)

    amedlat2 = amedlat - a_ML_static
    avert2 = avert - a_VT_static
    aantpost2 = aantpost - a_AP_static

    data = {'avert': avert2,
            'amedlat': amedlat2,
            'aantpost': aantpost2}
    df_corrected = pd.DataFrame(data)

    return df_corrected, avert2, amedlat2, aantpost2


import os
import numpy as np


def engine(root_folder, extension='.zoo', subfolders=None, name_contains=None,
           match_all=False, verbose=False):
    """
    Recursively search for files with a given extension, optionally filtering by
    specific subfolders and substrings in filenames.

    Arguments:
        root_folder (str): Root directory path where the search begins.
        extension (str): File extension to search for (e.g., '.zoo', '.c3d')
        subfolders (list or str, optional): Restrict search to folders with these names.
        name_contains (str or list, optional): Substring(s) that must appear in filename.
        match_all (bool, optional):
            If False, keep file if it contains ANY substring.
            If True, keep file only if it contains ALL substrings.
        verbose (bool): Print results.
    """

    # check format of subfolders
    if subfolders is not None:
        if isinstance(subfolders, str):
            subfolders = [subfolders]

    # check format of name_contains
    if name_contains is not None:
        if isinstance(name_contains, str):
            name_contains = [name_contains]

    matched_files = []
    subfolders_set = set(subfolders) if subfolders else None

    for dirpath, _, filenames in os.walk(root_folder):

        # Restrict to allowed subfolders
        if subfolders_set is not None:
            rel_path = os.path.relpath(dirpath, root_folder)
            if rel_path != '.':
                folder_parts = rel_path.split(os.sep)
                if not any(part in subfolders_set for part in folder_parts):
                    continue

        # Check each file
        for file in filenames:
            if not file.lower().endswith(extension.lower()):
                continue

            full_path = os.path.join(dirpath, file)

            # Substring filtering
            if name_contains is not None:
                file_lower = full_path.lower()
                checks = [(substr.lower() in file_lower) for substr in name_contains]

                if match_all and not all(checks):
                    continue
                if not match_all and not any(checks):
                    continue

            matched_files.append(full_path)

    # sort list
    matched_files = np.sort(matched_files)

    if verbose:
        print("Found {} {} file(s) in subfolder(s) {} with substrings {} (match_all={}):"
              .format(len(matched_files), extension, subfolders, name_contains, match_all))
        for f in matched_files:
            print("{}".format(f))

    return matched_files


if __name__ == '__main__':
    current_dir = os.path.dirname(os.path.abspath(__file__))
    project_root = os.path.dirname(os.path.dirname(os.path.dirname(current_dir)))
    sample_dir = os.path.join(project_root, 'data', 'sample_study', 'raw c3d files')

    # Example: match any of the substrings
    engine(sample_dir, extension='.c3d',
           subfolders=['Straight'],
           name_contains=['HC03', 'HC04'],
           match_all=False,
           verbose=True)

    # example patch match both strings
    c3d_files = engine(sample_dir, extension='.c3d',
                       subfolders=['Straight'],
                       name_contains=['HC03', '10'],
                       match_all=True,
                       verbose=True)
